!
! This example creates HDF5 file in a parallel environment
!

      PROGRAM FILEIO

      USE HDF5              ! This module contains all necessary modules 
      USE hdf_param
      IMPLICIT NONE

      INCLUDE 'mpif.h'

      real*4, allocatable, dimension(:) :: data

      INTEGER :: i,k1
      !


CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC


      comm = MPI_COMM_WORLD
      info = MPI_INFO_NULL

      CALL MPI_INIT(mpierror)
      CALL MPI_COMM_SIZE(comm, mpi_size, mpierror)
      CALL MPI_COMM_RANK(comm, mpi_rank, mpierror) 

C     Initialize the fortran interface for HDF5
      CALL h5open_f(error)
         
C     Setup file access property list with parallel I/O access,

      CALL h5pcreate_f(H5P_FILE_ACCESS_F,pcrp_list,error)
      
      CALL h5pset_fapl_mpio_f(pcrp_list,comm,info,error)


CCCC The no of procs has to exactly what was used to write the file
CCCC this part of the code checks if the no. of procs is ok or not.
CCCC This is done by checking the dimensions of the dataset 'LENGTHS'.

c     OPEN the input file to read.

      CALL h5fopen_f(pfilename,H5F_ACC_RDONLY_F,pfile_id,
     &     error,pcrp_list)
      if (error.ne.0) then
         write(*,*) 'File open of input file failed'
      endif

C     new code
      pgrpname = "/"
      pdsetname = "LENGTHS"
      CALL h5gopen_f(pfile_id,pgrpname,pgrp_id,error)
      CALL h5dopen_f(pgrp_id,pdsetname,pdset_id, error)
      if (error.ne.0) then
         write(*,*) 'dataset open failed'
      endif

C     Get dataset's dataspace handle.

      CALL h5dget_space_f(pdset_id,pdataspace,error)

C     Get dataspace's rank.
      CALL h5sget_simple_extent_dims_f(pdataspace,
     &     dimsw, maxdimsw, error)
      dim = dimsw(1)

      if(dim.eq.mpi_size) then 
C        write(*,*) 'good'
      else
         write(*,*) 'plz use:',mpi_size,'no. of procs'
         CALL MPI_finalize()
         CALL exit()
      endif
      
      CALL h5gclose_f(pgrp_id,error)
      CALL h5dclose_f(pdset_id,error)
      CALL h5sclose_f(pdataspace,error)
C new code

CCC~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
CCC~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
c     Create the op file collectively. The structure of the output file C has to be same as that of the original file. So we create the o/p 
C file here

      CALL h5fcreate_f(pfilename2,H5F_ACC_TRUNC_F,pfile_id2,
     &     error,access_prp = pcrp_list)
      
c     Close property list.
      CALL h5pclose_f(pcrp_list, error)

      
C     get the value of dim
      i =1
      write(pdsetname,1111) i
      write(pgrpname,2222) i

 1111 format('EigenVector',I2.2)
 2222 format('group',I2.2)
      CALL h5gopen_f(pfile_id,pgrpname,pgrp_id,error)
      CALL h5dopen_f(pgrp_id,pdsetname,pdset_id, error)
      if (error.ne.0) then
         write(*,*) 'dataset open failed'
      endif

C     Get dataset's dataspace handle.

      CALL h5dget_space_f(pdset_id,pdataspace,error)

C     Get dataspace's rank.
      CALL h5sget_simple_extent_dims_f(pdataspace,
     &     dimsw, maxdimsw, error)
      dim = dimsw(1)

      CALL h5sclose_f(pdataspace,error)
      CALL h5gclose_f(pgrp_id,error)
      CALL h5dclose_f(pdset_id,error)

CC      write(*,*) 'dims are:',dim

      do i=1,15
         dims(1) = dim
         
         maxdims = (/H5S_UNLIMITED_F/)
         
c     Create dataspace for the dataset.
         
         CALL h5screate_simple_f(rank,dims,pdataspace2,error,maxdims)
         CALL h5pcreate_f(H5P_DATASET_CREATE_F,pcrp_list,error)
            
c     Enable chunking in the dataset with chunk size of one vector
            
         chunk_dims(1) = dim
         CALL h5pset_chunk_f(pcrp_list,rank,chunk_dims,error)
            
c     Disabe filling of the dataset upon creation. By default
c     the dataset is filled with all zeroes.
            
         call h5pset_fill_time_f(pcrp_list,
     &        H5D_FILL_TIME_NEVER_F,error)
         
         write(pdsetname2,1111) i
         write(pgrpname2,2222) i
CC         write(*,*) pdsetname2
            
c     Now all setup has been completed, so create the dataset inside the HDF file
         CALL h5gcreate_f(pfile_id2,pgrpname2,pgrp_id2,error)            
         CALL h5dcreate_f(pgrp_id2,pdsetname2,H5T_NATIVE_REAL,
     &        pdataspace2,pdset_id2,error,pcrp_list)
            
         CALL h5pclose_f(pcrp_list,error)
         CALL h5sclose_f(pdataspace2,error)
         CALL h5gclose_f(pgrp_id2,error)
         CALL h5dclose_f(pdset_id2,error)
C         CALL h5fclose_f(pfile_id2,error)
      enddo
         CALL h5fclose_f(pfile_id2,error)
         CALL h5fclose_f(pfile_id,error)
         call h5close_f(error)


CCC File creation complete.....The actual IO part begins now.
CCC the code above this does not have any effect on performance(unless
CCC  i am missing something)

      parallelOffset = 0
      allocate(lenarray(mpi_size))
      pgrpname = '/'
      CALL phdfread(pfilename,pgrpname,ldsetname,
     &     lenarray,mpi_size,0,error)

      do i = 1,mpi_rank
         parallelOffset=parallelOffset+lenarray(i)
      enddo
      
      length(1) = lenarray(mpi_rank+1)

CCC At this point each process know how many elements it has to write
CCC And also the offset in file at which it has to write the data.
      deallocate(lenarray)

      allocate(data(length(1)))
      readsum = 0
      writesum = 0
      do k1 = 1,15
         
         write(pdsetname,1111) k1
         write(pdsetname2,1111) k1
         write(pgrpname,2222) k1
         write(pgrpname2,2222) k1
         
         do i=1,length(1)
            data(i) = mpi_rank
         enddo

C Read data from original file.
         start = MPI_WTIME()
         call phdfread(pfilename,pgrpname,pdsetname,
     &        data,length(1),parallelOffset,error)
         end = MPI_WTIME() - start
         CALL MPI_allreduce(end,endall,1,MPI_REAL8,
     &        MPI_MAX,comm,mpierror)
         readsum = readsum + endall


C Write data into the new o/p file.
         start = MPI_WTIME()
         call phdfwrite(pfilename2,pgrpname2,pdsetname2,
     &        data,length(1),parallelOffset,error)
         end = MPI_WTIME() - start
         CALL MPI_allreduce(end,endall,1,MPI_REAL8,
     &        MPI_MAX,comm,mpierror)

         writesum = writesum + endall
      enddo
      if(mpi_rank.eq.0) then
         write(*,*) 'Read sum is:',readsum
         write(*,*) 'Write sum is:',writesum
      endif

C      write(*,*) 'Read sum is:',readsum
C      write(*,*) 'Write sum is:',writesum

      deallocate(data)


      CALL MPI_FINALIZE(mpierror)
      if(mpi_rank.eq.0) call system("h5diff -v mfdn.smwf001 
     & hdfopfile > Differences.txt")
      END PROGRAM FILEIO
      
